package dhm 
{
	import com.efg.framework.CustomEventSound;
	import dhm.Contestant;
	import flash.display.*;
	import flash.filters.GlowFilter;
	import flash.geom.*;
	
	/**
	 * ...
	 * @author DHM
	 * 
	 * Class representing a flying disk weapon.
	 */
	public class Disk extends Weapon
	{
		private static const NORMAL_WIDTH:int = 8;
		private static const NORMAL_HEIGHT:int = 6;
		
		private var _inFlight:Boolean = false;
		private var _returning:Boolean = false;
		private var _nextTarget:Point;
		private var _speed:int;
		private var _velocity:Point = new Point(0, 0);
		private var _movesRemaining:int;
		private var _color:uint;
		private var _glowFilter:GlowFilter = new GlowFilter(0xFFFFFF, 0.8);

		public function Disk(color:uint, speed:Number)
		{
			super(1);
			_color = color;	
			_speed = speed;
			redraw();
		}
		
		public function get isInFlight():Boolean  { return _inFlight; }
		public function get effectiveSpeed():int { return _speed * boostFactor; }
		
		public function launch(target:Point):void
		{
			isBlocking = false;
			_inFlight = true;
			_armed = true;
			_returning = false;
			_nextTarget = target;
			plotTargetMoves();
			
			dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_THROW, false, 1, 0));
		}
		
		override public function reset():void
		{
			super.reset();
			_inFlight = false;
			_returning = false;
			rotation = 0;
			redraw();
		}
		
		override public function update():void
		{
			if (!_inFlight)
			{
				super.update();
				return;
			}
			
			if (_movesRemaining > 0) 
			{
				if (_returning)
					onReturningUpdate();
				_nextPosition.x = x + _velocity.x;
				_nextPosition.y = y + _velocity.y;
				_movesRemaining--;
			} 
			else 
			{
				// no moves remaining
				if (_returning)
				{	
					// returned to contestant
					onCatch();
				}
				else
				{
					// start return trip
					_returning = true;
					onReturningUpdate();
				}
			}
		}
		
		private function onCatch():void 
		{
			_inFlight = false;
			_returning = false;
			_armed = false;
			rotation = 0;
			redraw();
			dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_CATCH, false, 1, 0));	
		}
		
		override public function onBlocked():void 
		{
			super.onBlocked();
			_armed = false;
			_returning = true;
			onReturningUpdate();
			redraw();
		}
		
		protected override function redraw():void
		{
			var filters:Array = [];
			
			graphics.clear();
			graphics.lineStyle(1, 0xffffff);
			graphics.beginFill(_color);
			
			if (_inFlight && !_armed)
			{
				graphics.drawEllipse(0, 0, 2, 8);
			}
			else
			{
				if(isBlocking)
				{
					switch(facing)
					{
						case Facing.LEFT:
						case Facing.RIGHT:
							graphics.drawEllipse(0, 0, 8, 24);
							break;
						case Facing.DOWN:
							graphics.drawEllipse(0, 0, 20, 12);
							break;
						case Facing.UP:
							graphics.drawEllipse(0, 0, 20, 6);
							break;
					}
					
					filters.push(_glowFilter);
				}
				else
				{
					// draw normal
					graphics.drawEllipse(0, 0, NORMAL_WIDTH, NORMAL_HEIGHT);
				}
			}
				
			graphics.endFill();
			
			this.filters = filters;
		}
		
		private function plotTargetMoves():void
		{
			var distance:Number = GeometryUtility.distance1(_nextTarget, this);
			var effectiveSpeed:Number = _speed * (_armed ? boostFactor : Settings.BLOCKED_DISK_SPEED_FACTOR);
			_movesRemaining = Math.floor(Math.abs(distance / effectiveSpeed));
			_velocity.x = (_nextTarget.x - x) / _movesRemaining;
			_velocity.y = (_nextTarget.y - y) / _movesRemaining;
		}
		
		// update return path based on current home position
		private function onReturningUpdate():void 
		{
			if (!_armed)
				rotation += 20; // spin if disarmed
			
			if (_nextTarget.equals(_homePosition))
				return;
			_nextTarget.x = _homePosition.x;
			_nextTarget.y = _homePosition.y;
			plotTargetMoves();
		}
	}
}