package dhm 
{
	import com.efg.framework.CustomEventSound;
	import com.efg.framework.TileSheet;
	import flash.display.Bitmap;
	import flash.geom.*;
	
	/**
	 * ...
	 * @author DHM
	 * 
	 * Base enemy class.
	 */
	public class Enemy extends Contestant
	{
		private var _attackFrameCounter:int; // attack delay counter
		private var _minFramesBeforeAttack:int; // attack delay
		
		private var _dodging:Boolean;
		private var _dodgeFrameCounter:int; // dodge delay counter
		private var _minFramesBeforeDodge:int; // dodge delay
		
		private var _maneuverFrameCounter:int; // maneuver delay counter
		private var _minFramesBeforeManeuver:int; // maneuver delay
		
		private var _intelligence:int; // enemy intelligence rating
		
		protected var _maxPreferredAttackRange:int; // enemy will try to stay within this range of the player
		
		public function Enemy(tileImage:Bitmap, worldBounds:Rectangle, intelligence:int, speed:Number, hitPoints:int, weapon:Weapon)
		{
			super(tileImage, worldBounds, weapon, speed, hitPoints);
			
			_intelligence = intelligence;
			
			_minFramesBeforeManeuver = Math.floor(Settings.MAXIMUM_ENEMY_INTELLIGENCE / _intelligence) * Main.FRAME_RATE;
			_minFramesBeforeAttack = _minFramesBeforeManeuver;
			_minFramesBeforeDodge = _minFramesBeforeManeuver * 1.5;
			
			_attackFrameCounter = 0;
			_dodgeFrameCounter = _minFramesBeforeDodge;
			_maneuverFrameCounter = _minFramesBeforeManeuver;
			
			_maxPreferredAttackRange = 35000 / _intelligence;
		}
		
		public function get killPoints():int
		{
			return Math.ceil(_intelligence * Settings.ENEMY_INTELLIGENCE_KILL_POINT_RATIO);
		}

		override public function hit(damage:int):void 
		{
			super.hit(damage);
			dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_HIT_ENEMY, false, 1, 0));	
		}
		
		public function attack(player:Player):void {}
		
		public function maneuver(player:Contestant):void
		{
			// always avoid walls
			if (avoidWalls())
			{
				_maneuverFrameCounter = 0;
				_dodging = false; // dodging into a wall should be avoided
				return;
			}
			
			if (dodge(player))
			{
				_maneuverFrameCounter = 0;
				return;
			}
			
			if (_maneuverFrameCounter < _minFramesBeforeManeuver)
			{
				_maneuverFrameCounter++;
				//trace("Enemy - too soon to maneuver");
				return;
			}
			
			if (chase(player))
			{
				_maneuverFrameCounter = 0;
				return;
			}
			
			if (wander())
			{
				_maneuverFrameCounter = 0;
				return;
			}
		}
		
		private function avoidWalls():Boolean
		{
			var result:Boolean = false;
			var dx:int = velocity.x;
			var dy:int = velocity.y;
			
			if ((dx > 0 && _positionBounds.right - x < 2) ||
				(dx < 0 && x - _positionBounds.left < 2))
			{
				dx = -dx;
				result = true;
			}
			
			if ((dy > 0 && _positionBounds.bottom - y < 2) ||
				(dy < 0 && y - _positionBounds.top < 2))
			{
				dy = -dy;
				result = true;
			}
			
			if (result)
				updateDirection(dx, dy);
				
			return result;
		}
		
		private function wander():Boolean
		{
			if (!iqTest(0.2))
				return false;
				
			// randomly change direction
			var dx:int = velocity.x;
			var dy:int = velocity.y;
			
			dx = RandomUtility.choose(2) == 0 ? -1 : 1;
			dy = RandomUtility.choose(2) == 0 ? -1 : 1;
			
			//trace("Enemy - wandering");
			
			updateDirection(dx, dy);
			
			return true;
		}
		
		private function chase(player:Contestant):Boolean
		{
			if (!iqTest(0.4))
				return false;
			
			var dx:int = 0;
			var dy:int = 0;
			
			var distance:Number = GeometryUtility.distance1(player.center, this);
				
			// try to stay within attack radius
			if (distance > _maxPreferredAttackRange)
			{
				// move toward player
				dx = player.x - x;
				dy = player.y - y;
				//trace("Enemy - moving toward player");
			}
			else
			{
				//trace("Enemy - player within range");
				return false;
			}
			
			updateDirection(dx, dy);
			return true;
		}
		
		private function dodge(player:Contestant):Boolean
		{
			var playerWeapon:Weapon = player.weapon;
			
			if (_dodging)
			{
				if (!playerWeapon.isArmed || _dodgeFrameCounter > Settings.MAXIMUM_ENEMY_DODGE_FRAMES)
					// stop dodging
					_dodging = false;
				else
				{
					// continue dodging
					_dodgeFrameCounter++;
					//trace("Enemy - still dodging!");
					return true; // still dodging
				}
			}
			
			if (_dodgeFrameCounter < _minFramesBeforeDodge)
			{
				_dodgeFrameCounter++;
				//trace("Enemy - too soon to dodge!");
				return false;
			}
			
			// dodge only if an armed weapon is getting too close
			if (!playerWeapon.isArmed ||
				GeometryUtility.distance1(center, playerWeapon) > Settings.ENEMY_DODGE_RADIUS)
				return false;
				
			if (!iqTest(0.80))
			{
				//trace("Enemy - too dumb to dodge!");
				return false;
			}
			
			_dodging = true;
			_dodgeFrameCounter = 0;
			
			var dx:Number = x - playerWeapon.x;
			var dy:Number = y - playerWeapon.y;
			
			// 0=horizontal(-), 1=vertical(|), 2=diagonal(\), 3=diagonal(/)
			var dodgeDirection:int = 0;
			
			if (dx != 0)
			{
				var slope:Number = dy / dx;
			
				if (Math.abs(slope) > 0.75)
					dodgeDirection = 0;
				else if(Math.abs(slope) < 0.25)
					dodgeDirection = 1;
				else if (slope > 0)
					dodgeDirection =  2;
				else
					dodgeDirection = 3;
			}
				
			var dirX:int;
			var dirY:int;
			
			switch(dodgeDirection)
			{
				case 1:
					//trace("dodge (|)");
					dirX = 0;
					dirY = y > _positionBounds.height / 2 ? -1 : 1;
					break;
				case 2:
					//trace("diagonal dodge (/)");
					dirY = y > _positionBounds.height / 2 ? -1 : 1;
					dirX = -dirY;
					break;
				case 3:
					//trace("diagonal dodge (\\)");
					dirY = y > _positionBounds.height / 2 ? -1 : 1;
					dirX = dirY;
					break;
				default:
					//trace("dodge (-)");
					dirX = x > _positionBounds.width / 2 ? -1 : 1;
					dirY = 0;
					break;
			}
			
			updateDirection(dirX, dirY);
			
			return true;
		}
		
		// bar: be a value between 0 and 1: the higher the bar, the more difficult the test
		// cutoff: always returns false if intelligence is below the cutoff
		protected function iqTest(bar:Number, cutoff:int = 0):Boolean
		{
			if (_intelligence < cutoff)
				return false;
			return Math.random() * Settings.MAXIMUM_ENEMY_INTELLIGENCE * bar < _intelligence;
		}
		
		protected function checkAttackAbility():Boolean
		{
			if (_attackFrameCounter < _minFramesBeforeAttack)
			{
				_attackFrameCounter++;
				//trace("Enemy - too soon to attack!");
				return false;
			}
			
			if (!iqTest(0.8))
			{
				//trace("Enemy - too dumb to attack!");
				return false;
			}
			
			return true;
		}
		
		protected function onAttackStarted():void
		{
			_attackFrameCounter = 0;
		}
	}
}