package dhm 
{
	import flash.display.DisplayObject;
	import flash.geom.Point;
	/**
	 * ...
	 * @author DHM
	 */
	public final class GeometryUtility 
	{
		public static function distance1(p:Point, d:DisplayObject):Number
		{
			return distance2(p.x, d.x, p.y, d.y);
		}
		
		public static function distance2(x1:Number, x2:Number, y1:Number, y2:Number):Number
		{
			var dx:Number = x1 - x2;
			var dy:Number = y1 - y2;
			return Math.sqrt(dx * dx + dy * dy);
		}

		public static function distance3(p1:Point, p2:Point):Number
		{
			return distance2(p1.x, p2.x, p1.y, p2.y);
		}
		
		// Use quadratic formula to find an intercept point, if any exist.
		// target: current position of the target
		// targetVelocity: current velocity of the target
		// start: the starting position of the projectile
		// speed: the speed of the projectile
		// maxTime: the maximum allowable flight time for the projectile
		public static function intercept(target:Point, targetVelocity:Point, start:Point, speed:int, maxTime:int):Point
		{
			//trace("target=" + target);
			//trace("targetVelocity=" + targetVelocity);
			//trace("start=" + target);
			
			var dx:Number = target.x - start.x;
			var dy:Number = target.y - start.y;
			
			var a:Number = targetVelocity.x * targetVelocity.x + targetVelocity.y * targetVelocity.y - speed * speed;
			var b:Number = 2 * (targetVelocity.x * dx + targetVelocity.y * dy);
			var c:Number = dx * dx + dy * dy;
			
			var t:Number; // intercept time
			if (a != 0)
			{
				// quadratic
				var d:Number = b * b - 4 * a * c; // discriminant
				
				//trace("a=" + a + ", b=" + b + ", c=" + c + ", d=" + d);
				if (d < 0) 
				{
					//trace("no real solutions");
					return null;
				}
				
				var sqrtd:Number = Math.sqrt(d);
				t = (sqrtd - b) / (2 * a);
				if (t < 0 || t > maxTime)
					t = (-sqrtd - b) / (2 * a);
			}
			else if(b != 0)
			{
				// linear
				t = -(c / b);
			}
			else
			{
				// no solutions
				return null;
			}
			
			if (t < 0)
			{
				//trace("no positive solutions");
				return null;
			}
			
			if (t > maxTime)
			{
				//trace("maximum intercept time exceeded");
				return null;
			}
				
			var interceptPoint:Point = new Point(target.x + t * targetVelocity.x, target.y + t * targetVelocity.y);
			//trace(interceptPoint.toString());
			return interceptPoint;
		}
	}
}