package dhm 
{
	import dhm.Contestant;
	import dhm.Player;
	import flash.display.Bitmap;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	/**
	 * ...
	 * @author DHM
	 * 
	 * Guards carry stun poles instead of disks and they attack by charging the player.
	 */
	public class Guard extends Enemy
	{
		private var _charging:Boolean;
		private var _movesRemaining:int;
		
		public function Guard(tileImage:Bitmap, worldBounds:Rectangle, intelligence:int, speed:Number, hitPoints:int) 
		{
			super(tileImage, worldBounds, intelligence, Settings.PLAYER_MOVE_SPEED + 1, 1, new StunPole());
			_maxPreferredAttackRange = 150;
		}
		
		override public function attack(player:Player):void 
		{
			if (!weapon.isArmed)
				return;
				
			if (!checkAttackAbility())
				return;
			
			var distance:Number = GeometryUtility.distance1(player.center, weapon);
			if (distance > _maxPreferredAttackRange)
			{
				//trace("Guard - too far away to charge.");
				return;
			}
			
			charge(player);
			onAttackStarted();
		}

		override public function maneuver(player:Contestant):void 
		{
			if (_charging)
				return; // skip base class maneuver if charging
			super.maneuver(player);
		}
		
		override public function update():void
		{
			if (_charging)
			{
				if (_movesRemaining > 0) 
				{
					//trace("Guard - still charging");
					super.update();
					_movesRemaining--;
				}
				else
				{
					_charging = false;
					updateDirection(0, 0);
				}
			}
			
			super.update();
		}
		
		private function charge(player:Player):void
		{
			//trace("Guard - charging player");
			_charging = true;
			
			var target:Point = player.center;
			
			// we need to rewrite most of updateDirection since the charge can happen along any vector
			var distance:Number = GeometryUtility.distance1(target, weapon);
			var effectiveSpeed:Number = speed * Settings.CHARGE_SPEED_FACTOR;
			_movesRemaining = Math.floor(Math.abs(distance / effectiveSpeed));
			velocity.x = (target.x - weapon.x) / _movesRemaining;
			velocity.y = (target.y - weapon.y) / _movesRemaining;
			animationLoop = true;
			onVelocityChanged();
		}
	}
}