package dhm 
{
	import com.efg.framework.*;
	import flash.geom.*;
	import flash.events.*;
	import flash.text.*;
	import flash.ui.*;
	
	/**
	 * ...
	 * @author DHM
	 */
	public class Main extends GameFrameWork
	{
		//custom sccore board elements
		public static const SCORE_BOARD_LEVEL:String = "level";
		public static const SCORE_BOARD_SCORE:String = "score";
		public static const SCORE_BOARD_HEALTH:String = "health";
		public static const SCORE_BOARD_KILLS:String = "kills";
		
		public static const FRAME_RATE:int = 30;
		public static const GAME_WIDTH:int = 800;
		public static const GAME_HEIGHT:int = 640;
		
		private static const SCOREBOARD_HEIGHT:int = 30;
		private static const SCREEN_WIDTH:int = GAME_WIDTH;
		private static const SCREEN_HEIGHT:int = GAME_HEIGHT + SCOREBOARD_HEIGHT;
		
		private var _levelInCounter:int;
		private var _tronGame:TronArena;
		
		public function Main() 
		{
			if (stage) 
				addedToStage();
			else 
				addEventListener(Event.ADDED_TO_STAGE, addedToStage, false, 0, true);
		}
		
		override public function addedToStage(e:Event = null):void 
		{
			if (e != null)
				removeEventListener(Event.ADDED_TO_STAGE, addedToStage);

			super.addedToStage(e);
			
			addEventListener(Event.DEACTIVATE, deactivatedListener);
			
			init();
		}
		
		override public function init():void
		{		
			game = _tronGame = new TronArena();
			game.y = SCOREBOARD_HEIGHT;
			game.alpha = 0;
			setApplicationBackGround(SCREEN_WIDTH, SCREEN_HEIGHT, false, 0x000000);
			
			screenTextFormat = new TextFormat("_sans", "16", "0xffffff", "false");
			screenTextFormat.align = TextFormatAlign.CENTER;
			screenButtonFormat = new TextFormat("_sans", "12", "0x000000", "false");
			
			// paused screen
			pausedScreen = new BasicScreen(FrameWorkStates.STATE_SYSTEM_PAUSE, SCREEN_WIDTH, GAME_HEIGHT, true, 0x99000000);
			pausedScreen.y = SCOREBOARD_HEIGHT;
			pausedScreen.createDisplayText("Paused", SCREEN_WIDTH, new Point(0, 150), screenTextFormat);
			pausedScreen.createOkButton("Resume", new Point(SCREEN_WIDTH / 2 - 50, SCREEN_HEIGHT / 2), 100, 20, screenButtonFormat, 0x000000, 0x777777, 2);
			
			// scoreboard
			scoreBoard = new ScoreBoard();
			scoreBoard.mouseEnabled = false;
			addChild(scoreBoard);
			var scoreBoardLabelTextFormat:TextFormat = new TextFormat("_sans", "11", "0xffffff", "true");
			scoreBoardTextFormat= new TextFormat("_sans", "11", "0xffff66", "true");
			scoreBoard.createTextElement(SCORE_BOARD_LEVEL, new SideBySideScoreElement(80, 5, 20, "Level", scoreBoardLabelTextFormat, 25, "0", scoreBoardTextFormat));
			scoreBoard.createTextElement(SCORE_BOARD_SCORE, new SideBySideScoreElement(160, 5, 20, "Score", scoreBoardLabelTextFormat, 25, "0", scoreBoardTextFormat));
			scoreBoard.createTextElement(SCORE_BOARD_HEALTH, new SideBySideScoreElement(240, 5, 20, "Health", scoreBoardLabelTextFormat, 25, "0", scoreBoardTextFormat));
			scoreBoard.createTextElement(SCORE_BOARD_KILLS, new SideBySideScoreElement(320, 5, 20, "Kills", scoreBoardLabelTextFormat, 25, "0 / 0", scoreBoardTextFormat));
			
			// title screen
			titleScreen = new BasicScreen(FrameWorkStates.STATE_SYSTEM_TITLE, SCREEN_WIDTH, SCREEN_HEIGHT, false, 0x000000);
			titleScreen.createDisplayText("TRON: Arena", SCREEN_WIDTH, new Point(0, 150), screenTextFormat);
			titleScreen.createOkButton("Play", new Point(SCREEN_WIDTH / 2 - 50, SCREEN_HEIGHT / 2), 100, 20, screenButtonFormat, 0x000000, 0x777777, 2);
			
			// instructions screen
			instructionsScreen = new BasicScreen(FrameWorkStates.STATE_SYSTEM_INSTRUCTIONS, SCREEN_WIDTH, SCREEN_HEIGHT, false, 0x000000);
			instructionsScreen.createDisplayText("INSTRUCTIONS:\n\nArrow keys or A,D,W,S to move.\nSPACE to block.\nLeft-click mouse to throw.\n", SCREEN_WIDTH, new Point(0, 150), screenTextFormat);
			instructionsScreen.createOkButton("Start", new Point(SCREEN_WIDTH / 2 - 50, SCREEN_HEIGHT / 2), 100, 20, screenButtonFormat, 0x000000, 0x777777, 2);
			
			// level in screen
			levelInScreen = new BasicScreen(FrameWorkStates.STATE_SYSTEM_LEVEL_IN, SCREEN_WIDTH, GAME_HEIGHT, false, 0x000000);
			levelInScreen.y = SCOREBOARD_HEIGHT;
			levelInText = "Level ";
			levelInScreen.createDisplayText(levelInText, 100, new Point(SCREEN_WIDTH / 2 - 50, 150), screenTextFormat);
			
			// game over screen
			gameOverScreen = new BasicScreen(FrameWorkStates.STATE_SYSTEM_GAME_OVER, SCREEN_WIDTH, GAME_HEIGHT, false, 0x000000);
			gameOverScreen.y = SCOREBOARD_HEIGHT;
			gameOverScreen.createDisplayText("Game Over", 100, new Point(SCREEN_WIDTH / 2 - 50, 150), screenTextFormat);
			gameOverScreen.createOkButton("Restart", new Point(SCREEN_WIDTH / 2 - 50, 250), 100, 20, screenButtonFormat, 0x000000, 0x777777, 2);
			
			//set initial game state
			switchSystemState(FrameWorkStates.STATE_SYSTEM_TITLE);
			
			SoundLibrary.registerSounds(soundManager);
			
			//create timer and run it one time
			frameRate = FRAME_RATE;
			startTimer();
		}
		
		private var _lastPaused:Boolean = false;
		
		override public function systemGamePlay():void 
		{
			if (paused != _lastPaused)
			{
				_lastPaused = paused;
				_tronGame.interactive = !paused;
			}
			if(!paused)
				game.runGame();
		}
		
		override public function systemLevelIn():void
		{
			_levelInCounter = FRAME_RATE;
			levelInScreen.alpha = 1;
			game.alpha = 0;
			super.systemLevelIn();
		}
		
		override public function switchSystemState(stateval:int):void
		{
			super.switchSystemState(stateval);

			if(game.stage != null)
				_tronGame.interactive = !paused && currentSystemState == FrameWorkStates.STATE_SYSTEM_GAME_PLAY;
		}
		
		override public function systemWait():void
		{
			if (lastSystemState == FrameWorkStates.STATE_SYSTEM_LEVEL_IN) 
			{
				if (_levelInCounter > 0)
					_levelInCounter--;
				else
				{
					game.alpha = 1;
					levelInScreen.alpha -= 0.02;
					if (levelInScreen.alpha <= 0)
					{
						levelInScreen.alpha = 0;
						dispatchEvent(new Event(EVENT_WAIT_COMPLETE));
					}
				}
			}
		}
		
		private function deactivatedListener(e:Event):void
		{
			if (!paused && (currentSystemState == FrameWorkStates.STATE_SYSTEM_GAME_PLAY || currentSystemState == FrameWorkStates.STATE_SYSTEM_WAIT))
				pausedKeyPressedHandler();
		}
	}
}