package dhm 
{
	import com.efg.framework.*;
	import flash.display.*;
	import flash.events.*;
	import flash.filters.GlowFilter;
	import flash.geom.*;
	import flash.ui.*;
	import mx.rpc.soap.types.RowSetType;
	
	/**
	 * ...
	 * @author DHM
	 */
	public class TronArena extends Game
	{
		private static const MARGIN:int = 16;
		
		private var _backgroundBitmapData:BitmapData;
		private var _backgroundBitmap:Bitmap
		private var _currentLevel:int;
		private var _score:int;
		
		private var _keyPressList:Array = [];
		
		private var _crosshairs:Bitmap;
		private var _player:Player;
		private var _enemies:Array = [];
		private var _pickups:Array = [];
		
		private var _multipleKillBonusFactor:int = 1;
				
		private var _moveBounds:Rectangle;
		
		private var _weaponLayer:Layer;
		private var _contestantLayer:Layer;
		private var _pickupLayer:Layer;
		
		private var _addPickupFrameCounter:int;
		private var _maximumPickupsThisLevel:int;
		private var _actualPickupsThisLevel:int;

		private var _goalKillsThisLevel:int;
		private var _actualKillsThisLevel:int;
		private var _addEnemyFrameCounter:int;
		
		private var _enemyClass3SpawnChance:Number;
		private var _enemyClass2SpawnChance:Number;
		
		private var _enemySpawnPositions:Array = [];
		private var _pickupSpawnPositions:Array = [];
		
		private var _interactive:Boolean = false; // is the game accepting keyboard/mouse input?
		
		public function TronArena() 
		{
			// organize object classes into layers
			_weaponLayer = new Layer();
			_contestantLayer = new Layer();
			_pickupLayer = new Layer();
			
			_backgroundBitmapData = new BitmapData(Main.GAME_WIDTH, Main.GAME_HEIGHT, true, 0x00000000);
			_backgroundBitmap = new Bitmap(_backgroundBitmapData);
			
			addChild(_backgroundBitmap); // bottom layer
			addChild(_pickupLayer);
			addChild(_contestantLayer);
			addChild(_weaponLayer); // top layer
			
			_moveBounds = new Rectangle(MARGIN, MARGIN, Main.GAME_WIDTH - (2 * MARGIN), Main.GAME_HEIGHT - (MARGIN * 2));
			
			_enemySpawnPositions = [
				new Point(_moveBounds.left + MARGIN, _moveBounds.top), 
				new Point(_moveBounds.right - Contestant.WIDTH - MARGIN, _moveBounds.top), 
				new Point(_moveBounds.right - Contestant.WIDTH, _moveBounds.bottom - Contestant.HEIGHT - MARGIN),
				new Point(_moveBounds.left, _moveBounds.bottom - Contestant.HEIGHT - MARGIN)
			];
			
			_pickupSpawnPositions = [
				new Point(Main.GAME_WIDTH * 1 / 4, Main.GAME_HEIGHT / 2),
				new Point(Main.GAME_WIDTH * 3 / 4, Main.GAME_HEIGHT / 2),
			];
			
			focusRect = false;
			mouseChildren = false;
		}
		
		public function get interactive():Boolean { return _interactive; }
		
		public function set interactive(interactive:Boolean):void
		{
			if (_interactive == interactive)
				return;
				
			_interactive = interactive;
			
			//trace("interactive=" + _interactive);
			
			if (interactive)
			{
				// hide mouse and start capturing input
				Mouse.hide();
				
				_keyPressList = [];
				
				stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMoveListener);
				stage.addEventListener(MouseEvent.CLICK, mouseClickListener);
				stage.addEventListener(KeyboardEvent.KEY_DOWN, keyDownListener);
				stage.addEventListener(KeyboardEvent.KEY_UP, keyUpListener);
			}
			else
			{
				// show mouse and stop capturing input
				Mouse.show();
					
				stage.removeEventListener(MouseEvent.MOUSE_MOVE, mouseMoveListener);					
				stage.removeEventListener(MouseEvent.CLICK, mouseClickListener);
				stage.removeEventListener(KeyboardEvent.KEY_DOWN, keyDownListener);
				stage.removeEventListener(KeyboardEvent.KEY_UP, keyUpListener);
			}
		}
		
		override public function newGame():void 
		{
			drawLevelBackground();
			
			_currentLevel = 0;
			onCurrentLevelChanged();
			
			_score = 0;
			onScoreChanged();
			
			// create and center crosshairs
			_crosshairs = new ImageLibrary.CrosshairsGif();
			addChild(_crosshairs);
			_crosshairs.x = width / 2;
			_crosshairs.y = height / 2;
		}
		
		override public function newLevel():void
		{
			stage.focus = this;
			
			_currentLevel++;
			onCurrentLevelChanged();
			
			_actualPickupsThisLevel = 0;
			_maximumPickupsThisLevel = Math.max(Settings.MINIMUM_PICKUPS_PER_LEVEL, Math.min(Settings.MAXIMUM_PICKUPS_PER_LEVEL, _currentLevel));
			
			_goalKillsThisLevel = Math.ceil(Math.max(Settings.MINIMUM_ENEMIES_PER_LEVEL, _currentLevel / 2));
			_actualKillsThisLevel = 0;
			onKillsChanged();
			
			if(_player == null)
			{
				_player = new Player(_moveBounds);
				addContestant(_player);
			}
			else
			{
				_player.health += Settings.PLAYER_LEVEL_UP_BONUS_HIT_POINTS;
				_player.reset();
			}
			
			onHealthChanged();
			
			_player.moveTo(new Point(width / 2, height / 2));
			
			_enemyClass3SpawnChance = Math.min(0.6, (_currentLevel - 4) / 5);
			_enemyClass2SpawnChance = Math.min(0.8, (_currentLevel - 2) / 5);
		}
		
		override public function runGame():void 
		{
			if(_player.isAlive)
				_player.processKeboardInput(_keyPressList);
			
			checkAddPickup();
			checkAddEnemies();
			maneuver();
			update();
			checkCollisions();
			render();
			attack();
			checkForEndLevel();
			checkForEndGame();
		}
		
		private function checkAddEnemies():void
		{
			if (_enemies.length >= _enemySpawnPositions.length)
				return;
			
			var enemiesRemaining:int = _goalKillsThisLevel - _actualKillsThisLevel - _enemies.length;
			if (enemiesRemaining < 1)
				return;
				
			if (_addEnemyFrameCounter < Settings.MINIMUM_FRAMES_BEFORE_ADD_ENEMIES)
			{
				// special case -- if there are no enemies, don't wait more than a couple seconds before adding more
				_addEnemyFrameCounter++;
				if (_enemies.length == 0)
					_addEnemyFrameCounter = Math.max(Settings.MINIMUM_FRAMES_BEFORE_ADD_ENEMIES - Main.FRAME_RATE * 2, _addEnemyFrameCounter);
				return;
			}
			
			var newEnemyCount:int = Math.min(RandomUtility.choose(enemiesRemaining + 1), _enemySpawnPositions.length - _enemies.length);
			
			if (newEnemyCount > 0)
			{
				addEnemies(newEnemyCount);
				_addEnemyFrameCounter = 0;
			}
		}
		
		private function checkAddPickup():void
		{
			if (_maximumPickupsThisLevel <= _actualPickupsThisLevel)
				return;
				
			if (_pickups.length >= _pickupSpawnPositions.length)
				return;
				
			if (_addPickupFrameCounter < Settings.MINIMUM_FRAMES_BEFORE_ADD_PICKUPS)
			{
				_addPickupFrameCounter++;
				return;
			}
			
			if (Math.random() > 0.95)
			{
				_actualPickupsThisLevel++;
			
				var pickup:Pickup;
				switch(RandomUtility.choose(3))
				{
				case 0:
					pickup = new InvincibilityPickup();
					break;
				case 1:
					pickup = new HealthPickup();
					break;
				case 2:
					pickup = new SpeedPickup();
					break;
				}
				
				// find an empty pickup spawn position
				for (var i:int = 0; i < _pickupSpawnPositions.length; i++)
				{
					var available:Boolean = true;
					
					for each(var existingPickup:Pickup in _pickups)
					{
						if(existingPickup.spawnPointIndex == i)
							available = false;
					}
					
					if(available)
					{
						pickup.spawnPointIndex = i;					
						addPickup(pickup, _pickupSpawnPositions[i]);
						_addPickupFrameCounter = 0;
						break;
					}
				}
			}
		}
		
		// update every living contestant
		private function update():void
		{
			if(_player.isAlive)
				_player.update();
			
			var enemy:Enemy;
			
			for (var i:int = _enemies.length - 1; i >= 0; i--) 
			{					
				enemy = _enemies[i];
				
				if (enemy.shouldRemove)
				{
					removeContestant(enemy);
					_enemies.splice(i, 1);
					continue;
				}
				
				if (!enemy.isAlive)
					continue;
				
				enemy.update();
			}
		}
		
		private function checkCollisions():void
		{
			checkPlayerPickupCollisions();
			checkDiskWallCollisions();
			checkEnemyWeaponCollisions();
			checkPlayerWeaponCollisions();
		}
		
		private function checkPlayerPickupCollisions():void
		{
			var pickup:Pickup;
			
			for (var i:int = _pickups.length - 1; i >= 0; i--)
			{					
				pickup = _pickups[i];
				
				if (!_player.hitTestObject(pickup))
					continue;
				
				dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_PICKUP, false, 1, 0));

				pickup.Apply(_player);
				if (pickup is HealthPickup)
					onHealthChanged();
				
				removePickup(pickup);
				_pickups.splice(i, 1);
			}
		}
		
		// check for collisions between disk weapons and arena walls
		private function checkDiskWallCollisions():void
		{
			checkDiskWallCollision(_player.weapon as Disk);
			for each (var enemy:Enemy in _enemies) 
			{
				var disk:Disk = enemy.weapon as Disk;
				if(disk != null)
					checkDiskWallCollision(disk);
			}
		}
		
		private function checkDiskWallCollision(disk:Disk):void
		{
			if (!disk.isArmed)
				return;
				
			if (disk.x + disk.width > Main.GAME_WIDTH - 3 || disk.x < 3 || disk.y + disk.height > Main.GAME_HEIGHT - 3 || disk.y < 3)
				disk.onBlocked();
		}
		
		// check for collisions between the player's weapon and each enemy
		private function checkEnemyWeaponCollisions():void
		{
			var weapon:Weapon = _player.weapon;
			
			if (weapon.isArmed)
			{
				var enemy:Enemy;
			
				for (var i:int = _enemies.length - 1; i >= 0; i--) 
				{					
					enemy = _enemies[i];
					
					if (!enemy.isAlive)
						continue; // already dead
					
					checkContestantWeaponCollision(enemy, weapon);
					if (!enemy.isAlive)
					{
						_actualKillsThisLevel++;
						onKillsChanged();

						addToScore(enemy.killPoints * _multipleKillBonusFactor);
						_multipleKillBonusFactor *= Settings.MULTIPLE_KILL_BONUS_FACTOR;
					}
				}
			}
			else
			{
				// reset for the next throw
				_multipleKillBonusFactor = 1;
			}
		}
		
		// check for collisions between the player and each enemy's weapon
		private function checkPlayerWeaponCollisions():void
		{
			if (_player.isInvincible)
				return;
				
			for each(var enemy:Enemy in _enemies) 
			{		
				if (!_player.isAlive)
					break;
					
				if (checkContestantWeaponCollision(_player, enemy.weapon))
					onHealthChanged();
			}
		}
		
		private function checkContestantWeaponCollision(contestant:Contestant, weapon:Weapon):Boolean
		{
			if (contestant.isInvincible || !weapon.isArmed)
				return false;
				
			// see if the weapon is in the contestant's exclusion list
			var ignoreHitsIndex:int = contestant.hitTestExclusions.indexOf(weapon);
			
			if (ignoreHitsIndex < 0)
			{
				// included weapon
				
				if (contestant.weapon.isBlocking && contestant.weapon.hitTestObject(weapon))
				{
					// hit the contestant's blocking weapon
					weapon.onBlocked();
					return false;
				}
				else if (weapon.hitTest(contestant))
				{
					// hit the contestant
					weapon.onHit(contestant);
					// we only want to record damage on initial impact, not on every frame that ticks while the objects pass through one another
					contestant.hitTestExclusions.push(weapon);
					return true;
				}
			}
			else
			{
				// excluded weapon
				
				// If the objects are no longer touching, splice the ignored weapon from the exclusion list.
				// Note that we use a bounding box for hit testing in this case. Otherwise we could get two or more
				// hits in a row as the weapon passes through transparent areas of the sprite bitmap.
				// For example: hit first leg, between legs, hit second leg.  This seems unfair.
				if (!weapon.hitTestObject(contestant))
					contestant.hitTestExclusions.splice(ignoreHitsIndex, 1);
			}
			
			return false;
		}
		
		private function addEnemies(count:int):void
		{
			if (count > _enemySpawnPositions.length || count < 1)
				return; // bad argument
				
			// randomly place enemies at spawn points
			var remainingPositions:Array = [];
			for each(var p:Point in _enemySpawnPositions)
				remainingPositions.push(p);
				
			for (var i:int; i < count; i++)
			{
				var pIndex:int = RandomUtility.choose(remainingPositions.length);
				addEnemy(createRandomEnemy(), remainingPositions[pIndex]);
				remainingPositions.splice(pIndex, 1);
			}
		}
		
		private function createRandomEnemy():Enemy
		{
			var r:Number = Math.random();
			
			if (r < _enemyClass3SpawnChance)
			{
				if(RandomUtility.choose(2) == 0)
					return new Warrior(new ImageLibrary.Enemy3Png(), _moveBounds, 90, 3, 10, 2, 0xFFFF00);
				else
					return new Guard(new ImageLibrary.Enemy3Png(), _moveBounds, 75, 4, 1);
			}
			else if (r < _enemyClass2SpawnChance)
				return new Warrior(new ImageLibrary.Enemy2Png(), _moveBounds, 50, 4, 9, 1, 0xFF8800);
			else
				return new Warrior(new ImageLibrary.Enemy1Png(), _moveBounds, 30, 1.5, 12, 2, 0xFF0000);
		}
		
		private function addToScore(points:int):void
		{
			_score += points;
			onScoreChanged();
		}
		
		private function render():void
		{
			_player.render();
			
			for each(var enemy:Enemy in _enemies)
				enemy.render();
			
			for each(var pickup:Pickup in _pickups)
				pickup.render();
		}
		
		private function maneuver():void
		{
			if (!_player.isAlive)
				return;
			
			for each(var enemy:Enemy in _enemies)
			{
				if (!enemy.isAlive)
					continue;
				enemy.maneuver(_player);
			}
		}
		
		private function attack():void
		{
			if (!_player.isAlive || _player.isInvincible)
				return;
			
			for each(var enemy:Enemy in _enemies)
			{
				if (!enemy.isAlive)
					continue;
				enemy.attack(_player);
			}
		}
		
		private function checkForEndLevel():void 
		{
			// the level ends when the kill goal is reached
			if(_goalKillsThisLevel <= _actualKillsThisLevel && _enemies.length == 0)
			{
				// add level-up bonuses
				addToScore(_currentLevel * Settings.LEVEL_CLEAR_BONUS_POINT_RATIO);
				addToScore(_player.health * Settings.LEVEL_CLEAR_HEALTH_POINT_RATIO);
				
				cleanUpLevel();
				dispatchEvent(new Event(Game.NEW_LEVEL));
			}
		}
		
		private function checkForEndGame():void
		{
			if (!_player.shouldRemove)
				return; // player might be dead at this point, but still fading away
			
			cleanUpLevel();
			
			if(_player != null)
			{
				removeContestant(_player);
				_player = null;
			}
		
			dispatchEvent(new Event(Game.GAME_OVER));
			removeChild(_crosshairs);
			_crosshairs.bitmapData.dispose();
			_crosshairs = null;
		}
		
		private function cleanUpLevel():void
		{
			for each (var enemy:Enemy in _enemies)
				removeContestant(enemy);
			_enemies.splice(0, _enemies.length);
			
			for each (var pickup:Pickup in _pickups)
				removePickup(pickup);
			_pickups.splice(0, _pickups.length);
		}
		
		private function drawLevelBackground():void 
		{
			const backColor:uint = 0xFF1F1F1F;
			
			_backgroundBitmapData.lock();
			
			_backgroundBitmapData.fillRect(new Rectangle(0, 0, _backgroundBitmapData.width, _backgroundBitmapData.height), backColor);
			
			var currentShape:Shape = new Shape();
            currentShape.filters = [new GlowFilter(0x96C6FF, 0.8, 10, 10)];
			
			var graphics:Graphics = currentShape.graphics;
            graphics.lineStyle(1, 0x96C6FF);
			
			// border/walls
			graphics.moveTo(_moveBounds.left + 5, _moveBounds.y);
			graphics.lineTo(_moveBounds.left - 10, _moveBounds.bottom);
			graphics.lineTo(_moveBounds.right + 10, _moveBounds.bottom);
			graphics.lineTo(_moveBounds.right - 5, _moveBounds.y);
			graphics.lineTo(_moveBounds.left + 5, _moveBounds.y);
			
			// corners
			graphics.moveTo(_moveBounds.left + 5, _moveBounds.y);
			graphics.lineTo(_moveBounds.left - 3, 0);
			graphics.moveTo(_moveBounds.right - 5, _moveBounds.y);
			graphics.lineTo(_moveBounds.right + 3, 0);
			
			// grid lines
            graphics.moveTo(width / 4 + 5, _moveBounds.top);
            graphics.lineTo(width / 4 - 10, _moveBounds.bottom);
			graphics.moveTo(width / 2, _moveBounds.top);
            graphics.lineTo(width / 2, _moveBounds.bottom);
			graphics.moveTo(width * 3 / 4 - 5, _moveBounds.top);
            graphics.lineTo(width * 3 / 4 + 10, _moveBounds.bottom);
			
			// render pickup spawn positions as concentric ellipses
			var p:Point;
			for each(p in _pickupSpawnPositions)
			{
				var xOffsetDirection:int = p.x < Main.GAME_WIDTH / 2 ? -1 : 1;
				
				graphics.beginFill(backColor, 1);
				graphics.drawEllipse(p.x - 100, p.y - 80, 200, 160);
				graphics.endFill();
				graphics.beginFill(backColor, 1);
				graphics.drawEllipse(p.x - 50 + (xOffsetDirection * 2), p.y - 40 - 1, 100, 80);
				graphics.endFill();
				graphics.beginFill(backColor, 1);
				graphics.drawEllipse(p.x - 25 + (xOffsetDirection * 2), p.y - 20 - 2, 50, 40);
				graphics.endFill();
			}
				
			_backgroundBitmapData.draw(currentShape);
			
			// render enemy spawn position as glowing, red ellipses
			for each(p in _enemySpawnPositions)
			{
				currentShape = new Shape();
				graphics = currentShape.graphics;
				graphics.lineStyle(1, 0xFFFF2222, 0.5);
				graphics.drawEllipse(p.x + Contestant.WIDTH / 8, p.y + Contestant.HEIGHT / 2, Contestant.WIDTH * 3 / 4, Contestant.HEIGHT / 2);
				currentShape.filters = [new GlowFilter(0xFFFF0000)];
				_backgroundBitmapData.draw(currentShape);
			}
			
			_backgroundBitmapData.unlock();
		}
		
		private function addEnemy(enemy:Enemy, startPosition:Point):void 
		{
			enemy.moveTo(startPosition);
			_enemies.push(enemy);
			addContestant(enemy);
		}
		
		private function addPickup(pickup:Pickup, position:Point):void
		{
			pickup.moveTo(position);
			_pickups.push(pickup);
			_pickupLayer.addChild(pickup);
		}
		
		private function addContestant(contestant:Contestant):void
		{
			_contestantLayer.addChildAt(contestant, 0);
			_weaponLayer.addChild(contestant.weapon);
			
			contestant.addEventListener(CustomEventSound.PLAY_SOUND, soundEventRelay, false, 0, true);
			contestant.weapon.addEventListener(CustomEventSound.PLAY_SOUND, soundEventRelay, false, 0, true);
		}
		
		private function removeContestant(contestant:Contestant):void
		{
			contestant.removeEventListener(CustomEventSound.PLAY_SOUND, soundEventRelay);
			contestant.weapon.removeEventListener(CustomEventSound.PLAY_SOUND, soundEventRelay);
			
			_contestantLayer.removeChild(contestant);
			_weaponLayer.removeChild(contestant.weapon);
			contestant.dispose();
		}
		
		public function soundEventRelay(e:CustomEventSound):void 
		{
			dispatchEvent(e);
		}
		
		private function removePickup(pickup:Pickup):void
		{
			_pickupLayer.removeChild(pickup);
		}
		
		private function onHealthChanged():void
		{
			dispatchEvent(new CustomEventScoreBoardUpdate(CustomEventScoreBoardUpdate.UPDATE_TEXT, Main.SCORE_BOARD_HEALTH, String(_player.health)));
		}
		
		private function onCurrentLevelChanged():void
		{
			dispatchEvent(new CustomEventScoreBoardUpdate(CustomEventScoreBoardUpdate.UPDATE_TEXT, Main.SCORE_BOARD_LEVEL, String(_currentLevel)));
			dispatchEvent(new CustomEventLevelScreenUpdate(CustomEventLevelScreenUpdate.UPDATE_TEXT, String(_currentLevel)));
		}
		
		private function onScoreChanged():void
		{
			dispatchEvent(new CustomEventScoreBoardUpdate(CustomEventScoreBoardUpdate.UPDATE_TEXT, Main.SCORE_BOARD_SCORE, String(_score)));
		}
		
		private function onKillsChanged():void
		{
			dispatchEvent(new CustomEventScoreBoardUpdate(CustomEventScoreBoardUpdate.UPDATE_TEXT, Main.SCORE_BOARD_KILLS, String(_actualKillsThisLevel) + " / " + String(_goalKillsThisLevel)));
		}
		
		private function keyDownListener(e:KeyboardEvent):void
		{
			_keyPressList[e.keyCode] = true;
		}
		
		private function keyUpListener(e:KeyboardEvent):void
		{
			_keyPressList[e.keyCode] = false;
		}
		
		private function mouseClickListener(e:MouseEvent):void
		{
			if (mouseX < 0 || mouseY < 0)
				return;
			_player.throwWeapon(new Point(mouseX, mouseY));
		}
		
		private function mouseMoveListener(e:MouseEvent):void
		{
			_crosshairs.x = Math.max(0, mouseX - (_crosshairs.width / 2));
			_crosshairs.y = Math.max(0, mouseY - (_crosshairs.height / 2));
		}	
	}
}