package dhm 
{
	import com.efg.framework.CustomEventSound;
	import dhm.Player;
	import flash.display.Bitmap;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.sampler.NewObjectSample;
	/**
	 * ...
	 * @author DHM
	 * 
	 * An enemy armed with a flying disk.
	 */
	public class Warrior extends Enemy
	{
		private var _disk:Disk;
		
		public function Warrior(tileImage:Bitmap, worldBounds:Rectangle, intelligence:int, speed:Number, throwSpeed:Number, hitPoints:int, diskColor:uint) 
		{
			_disk = new Disk(diskColor, throwSpeed);
			super(tileImage, worldBounds, intelligence, speed, hitPoints, _disk);
		}
		
		override public function attack(player:Player):void 
		{
			if (_disk.isInFlight || _disk.isBlocking)
				return;
				
			if (!checkAttackAbility())
				return;
			
			var target:Point = calculateThrow(player);
			if(target != null)
			{
				_disk.launch(target);
				dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_THROW, false, 1, 0));
				onAttackStarted();
			}
		}
		
		private function calculateThrow(player:Contestant):Point
		{
			var target:Point = player.center;
			
			var overthrowMax:Number = 0;
			
			if (iqTest(0.9, 60))
			{
				// exact intercept calculation
				target = GeometryUtility.intercept(	target, 
													player.velocity, 
													new Point(_disk.x, _disk.y), 
													_disk.effectiveSpeed, 
													Settings.MAXIMUM_ENEMY_WEAPON_INTERCEPT_FRAMES);
				
				overthrowMax = 0.2;
			}
			else if (iqTest(0.6, 30))
			{
				// approximate intercept calculation:
				// 1. calculated the travel time to the target at its current location
				// 2. project the target's location after the calculated travel time based on its current velocity
				// 3. throw at the projected location
				var distance:Number = GeometryUtility.distance1(target, this);
				var timeEstimate:Number = distance / _disk.effectiveSpeed;
				var playerVelocity:Point = player.velocity;
				
				target.x += (playerVelocity.x * timeEstimate);
				target.y += (playerVelocity.y * timeEstimate);
				
				overthrowMax = 0.3;
			}
			else if (iqTest(0.2))
			{
				// throw directly at the target's current position
				overthrowMax = 0.2;
			}
			else
			{
				// add some random inaccuracy
				target.x += (RandomUtility.choose(2) == 0 ? 1 : -1) * Math.random() * 100;
				target.y += (RandomUtility.choose(2) == 0 ? 1 : -1) * Math.random() * 100;
			}
			
			if (overthrowMax > 0 && target != null)
			{
				// throw a bit beyond the target so that there is a chance of a return hit
				var overthrowFactor:Number = Math.min(overthrowMax, Math.random());
				
				target.x += overthrowFactor * (target.x - _disk.x);
				target.y += overthrowFactor * (target.y - _disk.y);
			}
			
			return target;
		}
	}
}