package dhm 
{
	import com.efg.framework.CustomEventSound;
	import flash.display.*;
	import flash.geom.*;
	
	/**
	 * ...
	 * @author DHM
	 * 
	 * Base class for weapons carried by contestants.
	 */
	public class Weapon extends Sprite
	{
		private var _damage:int;
		private var _facing:int = Facing.NONE;
		private var _blocking:Boolean = false; // is the weapon in blocking mode?
		
		protected var _armed:Boolean;
		
		protected var _nextPosition:Point = new Point();
		protected var _homePosition:Point = new Point();
		
		public var boostFactor:Number = 1; // speed boost multiplier -- only applies to thrown weapons
		
		public function Weapon(damage:int = 1)
		{
			_damage = damage;
		}
		
		public function get isArmed():Boolean { return _armed; }
		
		public function get isBlocking():Boolean { return _blocking; }
		public function set isBlocking(value:Boolean):void
		{
			if (_blocking == value)
				return;
			_blocking = value;
			onBlockingChanged();
		}
		
		public function get facing():int { return _facing; }
		public function set facing(value:int):void
		{
			if (_facing == value)
				return;
			_facing = value;
			onFacingChanged();
		}
		
		public function reset():void
		{
			_armed = false;
			boostFactor = 1;
			isBlocking = false;
		}	
		
		public function moveTo(position:Point):void
		{
			_nextPosition.x = x = position.x;
			_nextPosition.y = y = position.y;
			updateHomePosition(position);
		}
		
		public function updateHomePosition(position:Point):void
		{
			_homePosition.x = position.x;
			_homePosition.y = position.y;
		}
		
		public function update():void 
		{
			_nextPosition.x = _homePosition.x;
			_nextPosition.y = _homePosition.y;
		}
		
		public function render():void 
		{
			x = Math.max(0, _nextPosition.x);
			y = Math.max(0, _nextPosition.y);
		}
		
		public function hitTest(contestant:Contestant):Boolean
		{
			var p:Point = new Point(contestant.x, contestant.y);
			
			// we test against the weapon's bounding box since it's a close-enough approximation
			return contestant.hitTestBitmap(p, new Rectangle(x, y, width, height));
		}
		
		public function onHit(contestant:Contestant):void
		{
			contestant.hit(_damage);
		}
		
		public function onBlocked():void 
		{
			dispatchEvent(new CustomEventSound(CustomEventSound.PLAY_SOUND, SoundLibrary.SOUND_BLOCK, false, 1, 0));	
		}
		
		protected function redraw():void { }
		
		protected function onFacingChanged():void 
		{
			redraw();
		}
		
		protected function onBlockingChanged():void 
		{
			redraw();
		}
	}
}